;; Generated package description from cpupower.el  -*- no-byte-compile: t -*-
(define-package "cpupower" "1.0.4.0.20220506.173232" "cpupower command interface" 'nil :commit "518fa57d75a9199c1e779793599fea9f7f3f837f" :authors '(("Stephen Meister" . "pallagun@gmail.com")) :maintainer '("Stephen Meister" . "pallagun@gmail.com") :keywords '("hardware" "cpupower" "cpu" "frequency-scaling") :url "https://gitlab.com/steve-emacs-stuff/cpupower-el")
